module JCAB.HttpServer (
	StringAction, FileAction, EndAction,
	Context (..),
	listenLoop,
	doConnection
) where

import JCAB.String (padLeftTo, stripSpaces)
import JCAB.Regex (mkRegexNocase)
import JCAB.IO (fromIO)
import JCAB.Exception (throwDynIO)

import Data.Typeable (Typeable(..), mkTyConApp, mkTyCon)
import Control.Exception (handle, handleJust, ioErrors)
import qualified Control.Exception as X
import Control.Concurrent (forkIO)
import Data.Dynamic (fromDynamic)

import Network (Socket, accept)
import Network.URI (URI, uriPath, nullURI, parseURIReference)
import Text.Regex (matchRegex)
import Text.Html
import System.Locale (defaultTimeLocale)
import System.Time (getClockTime, toCalendarTime, formatCalendarTime)
import System.IO (Handle, hGetLine, hPutStr, hFlush, hClose, hSetBuffering, BufferMode(..))
import System.IO.Error (isEOFError, ioError)


type StringAction = IO String
type FileAction = Handle -> IO ()
type EndAction = IO ()

class Context c where
	logOut        :: c -> String -> IO ()
	addLogFilter  :: c -> (String -> StringAction) -> c
	processGetURI :: c -> URI -> IO (Maybe (Integer, String, StringAction, FileAction, EndAction))
--	processPutURI :: c -> URI -> IO (Maybe (Integer, FileAction))


listenLoop   :: Context c => Socket -> c -> IO ()
doConnection :: Context c => Handle -> c -> IO ()


-- Loops forever answering connection requests, and spawning threads to handle the connections.

listenLoop socket context = do
		logOut context $ "Listening on port 8000..."
		loop 0 -- We keep a connection counter, starting with zero.
	where
	loop connectionNum = do
		(h, name, number) <- accept socket
		hSetBuffering h LineBuffering
		logOut context $ "Connection received from " ++ name ++ ":" ++ show number

		-- Good logs must include useful information.
		let markLog s = do
			clockTime <- getClockTime
			calendarTime <- toCalendarTime clockTime
			let dateTime = formatCalendarTime defaultTimeLocale "(%Y/%m/%d %H:%M:%S)" calendarTime

			return $
				"[" ++ padLeftTo 6 (show connectionNum) ++ " " ++
				dateTime ++ " " ++ name ++ ":" ++ show number ++ "] " ++
				s

		-- This is the context passed in to the server's main function.
		-- It implements the customizable functionality.
		let connectionContext = addLogFilter context markLog

		forkIO $ do
			handle ( \e -> do
					logOut connectionContext $ "Exception in the HTTP server connection: " ++ show e
					return ()
				) $
				doConnection h connectionContext
			-- Close the handle when we're done with it.
			hFlush h
			hClose h

		-- Do again.
		loop (connectionNum+1)


doConnection h context = do

	logOut context "Started."

	handle ( \e -> logOut context $ "HttpServer: Unhandled exception: " ++ show e) $
		handleJust exceptions ( \e -> case e of
			InvalidRequestException line -> sendErrorResponse context h 400 $ "Bad request: '" ++ line ++ "'"
			InvalidURIException     uri  -> sendErrorResponse context h 400 $ "Bad URI: '" ++ uri ++ "'"
			ConnectionLostException      -> logOut context "HttpServer: Connection lost"
		) $ do
		(req, uri) <- receiveRequest h (logOut context)
		case req of
			"GET"  -> doGetRequest h context uri
			"HEAD" -> doHeadRequest h context uri
			otherwise -> do
				logOut context ("Unknown '" ++ req ++ "' request received for '" ++ show uri ++ "'")
				sendErrorResponse context h 501 "Not Implemented"

	logOut context "Finished."


data Exception
	= InvalidRequestException String
	| InvalidURIException     String
	| ConnectionLostException

instance Show Exception where
	show (InvalidRequestException s) = "HttpServer.InvalidRequestException " ++ s
	show (InvalidURIException     s) = "HttpServer.InvalidURIException "     ++ s
	show  ConnectionLostException    = "HttpServer.ConnectionLostException"

instance Typeable Exception where
	typeOf e = mkTyConApp (mkTyCon "HttpServer.Exception") []

exceptions :: X.Exception -> Maybe Exception
exceptions (X.DynException d) = fromDynamic d
exceptions _ = Nothing


doGetRequest h context uri = do
	logOut context ("GET Request received for '" ++ show uri ++ "'")
	result <- processGetURI context uri
	logOut context "Request processed"
	case result of
		Just (size, ctype, _, copyAction, endAction) -> do
			hPutStr h $ "HTTP/1.0 200 OK" ++ crlf ++
				"Content-Length: " ++ show size ++ crlf ++
				"Content-Type: " ++ ctype ++ crlf  ++
				crlf
			logOut context "Copy begins"
			copyAction h
			logOut context "Copy ended"
			endAction
			logOut context "EndAction ended"
		Nothing -> do
			logOut context $ "File not found: '" ++ show uri ++ "'"
			sendErrorResponse context h 404 "Not Found"


doHeadRequest h context uri = do
	logOut context ("HEAD Request received for '" ++ show uri ++ "'")
	result <- processGetURI context uri
	case result of
		Just (size, ctype, _, _, endAction) -> do
			hPutStr h $ "HTTP/1.0 200 OK" ++ crlf ++
				"Content-Length: " ++ show size ++ crlf ++
				"Content-Type: " ++ ctype ++ crlf  ++
				crlf
			endAction
		Nothing -> do
			logOut context $ "File not found: '" ++ show uri ++ "'"
			sendErrorResponse context h 404 "Not Found"


httpSeparators = "[()<>@,;:\\\"/?={} \t" -- "
httpToken      = "[^ \t]+" -- "[^" ++ httpSeparators ++ "]+"

requestRegex = mkRegexNocase $ "^(" ++ httpToken ++ ") ([^ \t]+|\\*) *(.*)?"

--- getRequestAcceptRegex			= myMkRegex "^Accept: .*"
--- getRequestAcceptLanguageRegex	= myMkRegex "^Accept-Language: .*"
--- getRequestAcceptEncodingRegex	= myMkRegex "^Accept-Encoding: .*"
--- getRequestUserAgentRegex		= myMkRegex "^User-Agent: .*"
--- getRequestHostRegex				= myMkRegex "^Host: .*"
--- getRequestConnectionRegex		= myMkRegex "^Connection:(.*)"



receiveRequest h writeLog = do
	line <-
		handleJust ioErrors ( \e -> case e of
			e | isEOFError e -> throwDynIO ConnectionLostException
			otherwise        -> ioError e
		) $
		fromIO stripSpaces $ hGetLine h

	if line == "" then
			-- Skipping any empty lines before the request.
			receiveRequest h writeLog
		else do
			case matchRegex requestRegex line of
				Just (req:newUri:_:[]) | newUri == "*" ->
					return (req, nullURI)
				Just (req:newUri:_:[]) ->
					case parseURIReference newUri of
						Just parsedURI | take 1 (uriPath parsedURI) == "/" ->
							return (req, parsedURI)
						otherwise -> do
							throwDynIO $ InvalidURIException newUri
				otherwise -> do
					throwDynIO $ InvalidRequestException line

crlf = "\r\n"

withAttrs htmlGen attrs h = htmlGen h ! attrs

sendErrorResponse context h code message = do
	logOut context $ "Error: " ++ show code ++ " " ++ message
	let htmlMessage = renderHtml htmlError
	let htmlSize = length htmlMessage
	hPutStr h $
		"HTTP/1.0 " ++ show code ++ " " ++ message ++ crlf ++
		"Content-Length: " ++ show htmlSize ++ crlf ++
		"Content-Type: text/html" ++ crlf  ++
		crlf ++
		renderHtml htmlError
	where
		htmlError =
			(header $
				thetitle $
					stringToHtml $ show code ++ " " ++ message) +++
			(withAttrs body [bgcolor "#cc9999"] $
				(h4 $ stringToHtml $ show code ++ " " ++ message) +++
					stringToHtml "No can do, sorry.")
