module JCAB.SSI(
	Context(..),
	parse,
) where

import           Prelude hiding(readFile)                    -- We're defining our own 'readFile' (callback) function here.
import           Text.ParserCombinators.Parsec hiding(parse) -- We're defining our own 'parse' function here.
import qualified Text.ParserCombinators.Parsec as Parsec     -- 'parse' will be qualified
import           Data.List(sort)
import qualified Data.HashTable as HT
import           Data.IORef


-- Module interface

class Context c where
	lookupVar   :: c -> String -> IO String -- var name -> value
	readFile    :: c -> String -> IO String -- fname -> file contents
	fileLastMod :: c -> String -> IO String -- fname -> file modification time

parse :: Context c => c -> String -> String -> IO (Either String String)
parse context fname file =
	case Parsec.parse ssiParser fname file of
		Left  err -> (return . Left . show) err
		Right p   -> do
			parsed <- ssiProcessFile context p
			return $ Right parsed


-- Custom parsec extensions

manyTillEnd :: GenParser tok st a -> GenParser tok st end -> GenParser tok st ([a], end)
manyTillEnd p end = scan
	where
	scan =
			do { endr <- end; return ([], endr) }
		<|>	do { x <- p; (xs, endr) <- scan; return (x:xs, endr) }

many1Till :: GenParser tok st a -> GenParser tok st end -> GenParser tok st [a]
many1Till p end = do { x <- p; xs <- manyTill p end; return (x:xs) }

many1TillEnd :: GenParser tok st a -> GenParser tok st end -> GenParser tok st ([a], end)
many1TillEnd p end = do { x <- p; (xs, endr) <- manyTillEnd p end; return (x:xs, endr) }

notFollowedBy_ :: Show a => GenParser tok st a -> GenParser tok st ()
notFollowedBy_ p = try $
		do	{ c <- p; unexpected (show [c]) }
	<|>	do	return ()


-- SSI file parser

type SsiDirective = (String, [(String, String)])
type SsiToken     = (String, Maybe SsiDirective)

ssiParser :: Parser [SsiToken]
ssiParser = do
	tresult <- ssiTextParser
	case tresult of
		(""  , True ) -> return []
		(text, True ) -> return [(text, Nothing)]
		(text, False) -> do
			(comment, dir) <- ssiDirectiveParser <|> ssiCommentParser
			rest <- ssiParser
			case (dir, rest) of
				(Nothing, (text2, dir2):rest2) -> return $ (text ++ comment ++ text2, dir2) : rest2
				otherwise                      -> return $ (text ++ comment         , dir ) : rest

ssiTextParser = manyTillEnd anyChar ssiEndOfTextParser

ssiEndOfTextParser = do
	(eof >> return True) <|> (try (string "<!--") >> return False)

ssiDirectiveParser =
	do	char '#'
		many space
		name <- many1 letter <?> "directive name"
		many space
		argList <- many ssiDirectiveArgumentParser
		try (string "-->")
		return ("", Just (name, (sort argList)))

ssiCommentParser = do
	cmt <- manyTill anyChar (try $ string "-->")
	return ("<!--" ++ cmt ++ "-->", Nothing)

ssiDirectiveArgumentParser = do
	name <- many1 letter
	char '='
	char '"'
	arg <- manyTill ssiArgStringCharParser $ char '"'
	many space
	return (name, arg)

ssiArgStringCharParser =
		(string "\\\"" >> return '"') -- "
	<|>	(anyChar)


-- Condition parser

parseCond :: String -> Either ParseError Bool
parseCond "" = Right False
parseCond s = Parsec.parse condition ("<<" ++ s ++ ">>") s

condition :: Parser Bool
condition =
		do	b <- try (spaces >> condExpr)
			(eof >> return b) <|> return True
	<|>	do	eof
			return False
	<|>	do	return True

condExpr =
		do	char '('
			spaces
			result <- condExpr
			char ')'
			spaces
			return result
	<|>	do	char '!'
			spaces
			result <- condExpr
			return $ not result
	<|>	do	condCompareStrings >>= doCond
	where
	doCond b1 =
			do	try (string "&&")
				spaces
				b2 <- condExpr
				return $ b1 && b2
		<|>	do	try (string "||")
				spaces
				b2 <- condExpr
				return $ b1 || b2

condCompareStrings = condJuxtaposition >>= doCond
	where
	doCond str1 =
			do	try (string "=")
				spaces
				str2 <- condJuxtaposition <|> return ""
				return $ str1 == str2
		<|>	do	try (string "!=")
				spaces
				str2 <- condJuxtaposition <|> return ""
				return $ str1 /= str2
		<|>	do	try (string "<=")
				spaces
				str2 <- condJuxtaposition <|> return ""
				return $ str1 <= str2
		<|>	do	try (string ">=")
				spaces
				str2 <- condJuxtaposition <|> return ""
				return $ str1 >= str2
		<|>	do	string "<"
				spaces
				str2 <- condJuxtaposition <|> return ""
				return $ str1 < str2
		<|>	do	string ">"
				spaces
				str2 <- condJuxtaposition <|> return ""
				return $ str1 > str2
		<|>	do	return $ str1 /= ""

condJuxtaposition = condToken >>= doCond
	where
	doCond tok1 =
			do	tok2 <- condJuxtaposition
				return $ tok1 ++ " " ++ tok2
		<|>	do	return tok1

condToken :: Parser String
condToken =
		do	char '"'
			token <- manyTill anyChar (char '"')
			spaces
			return token
	<|>	do	char '\''
			token <- manyTill anyChar (char '\'')
			spaces
			return token
	<|>	do	token <- many1Till anyChar ((space >> return ()) <|> eof <|> (lookAhead condOperator >> return ()))
			spaces
			return token

condOperator =
		(oneOf "=<>)" >> return "")
	<|>	string "!="
	<|>	string "&&"
	<|>	string "||"


-- Directive processing

data Context c => SsiContext c = SsiContext {
		ifTrue      :: IORef Bool,
		ifTrueStack :: IORef [Bool],
		varMap      :: HT.HashTable String String,
		external    :: c
	}

updateVar :: Context c => SsiContext c -> String -> String -> IO ()
updateVar c name value = HT.update (varMap c) name value >> return ()

instance Context c => Context (SsiContext c) where
	lookupVar c name = do
		value <- HT.lookup (varMap c) name
		case value of
			Just v    -> return v
			otherwise -> lookupVar (external c) name
	readFile    c = readFile    (external c)
	fileLastMod c = fileLastMod (external c)

ssiProcessFile :: Context c => c -> [SsiToken] -> IO String
ssiProcessFile context file = do
	ifTrueVar      <- newIORef True
	ifTrueStackVar <- newIORef []
	varMapVar      <- HT.new (==) HT.hashString

	let newContext = SsiContext {
			ifTrue      = ifTrueVar,
			ifTrueStack = ifTrueStackVar,
			varMap      = varMapVar,
			external    = context
		}
	processFile newContext file

processFile context [] = return ""
processFile context ((t, Nothing):rest) = do
	trest <- processFile context rest
	return $ t ++ trest
processFile context ((t, Just (name, argList)):rest) = do
	oldIfTrue <- readIORef (ifTrue context)
	subArgList <- sequence $ map (\(n, v) -> ssiSubstituteString context v >>= \x -> return (n, x)) argList
	td <- runDirective context name subArgList
	trest <- processFile context rest
	return $ (if oldIfTrue then t else "") ++ td ++ trest

ssiSubstituteString context str = do
	case Parsec.parse subParser "SSIargument" str of
		Left err -> return str
		Right s  -> process s
	where
	process [] = return ""
	process ((text, ""):rest) = do
		prest <- process rest
		return $ text ++ prest
	process ((text, name):rest) = do
		value <- lookupVar context name
		prest <- process rest
		return $ text ++ value ++ prest
	subParser = do
		r1 <- many $ try $ manyTillEnd anyChar (try varParser)
		t2 <- many anyChar
		return $ r1 ++ [(t2, "")]
	varParser = do
		char '$'
		varBracedName <|> varName
	varBracedName = do
		char '{'
		name <- varName
		char '}'
		return name
	varName = many1 (alphaNum <|> char '_')

doDirectiveIfTrue context d = do
	oldIfTrue <- readIORef (ifTrue context)
	case oldIfTrue of
		False -> return ""
		True  -> d

runDirective context "include" [("virtual", virtual)] = do
	doDirectiveIfTrue context $ do
		file <- readFile context virtual
		case Parsec.parse ssiParser virtual file of
			Left  err -> return $ " {SSI include '" ++ virtual ++ "' failed to parse file. Error message follows}\n" ++ show err
			Right p   -> do
				processFile context p
runDirective context "include" argList = return $ " {SSI 'include' directive has bad arguments: " ++ show argList ++ "} "

runDirective context "flastmod" [("virtual", virtual)] = doDirectiveIfTrue context $ fileLastMod context virtual
runDirective context "flastmod" argList = return $ " {SSI 'flastmod' directive has bad arguments: " ++ show argList ++ "} "

runDirective context "set" [("value", value), ("var", name)] = doDirectiveIfTrue context $ updateVar context name value >> return ""
runDirective context "set" argList = return $ " {SSI 'set' directive has bad arguments: " ++ show argList ++ "} "

runDirective context "echo" [("var", name)] = doDirectiveIfTrue context $ lookupVar context name
runDirective context "echo" argList = return $ " {SSI 'echo' directive has bad arguments: " ++ show argList ++ "} "

runDirective context "if" [("expr", cond)] =
	case parseCond cond of
		Right condResult ->
			do	oldIfTrue      <- readIORef (ifTrue      context)
				oldIfTrueStack <- readIORef (ifTrueStack context)
				writeIORef (ifTrue context) condResult
				writeIORef (ifTrueStack context) (oldIfTrue:oldIfTrueStack)
				return ""
		Left err ->
			do	oldIfTrue      <- readIORef (ifTrue      context)
				oldIfTrueStack <- readIORef (ifTrueStack context)
				writeIORef (ifTrue context) True
				writeIORef (ifTrueStack context) (oldIfTrue:oldIfTrueStack)
				return $ " {SSI 'if' condition is bad: '" ++ cond ++ "' - " ++ show err ++ "} "
runDirective context "if" argList = return $ " {SSI 'if' directive has bad arguments: " ++ show argList ++ "} "

runDirective context "elif" [("expr", cond)] =
	do	oldIfTrue <- readIORef (ifTrue context)
		if oldIfTrue then
				do	writeIORef (ifTrue context) False
					return ""
			else
				do	case parseCond cond of
						Right False ->
							do	return ""
						Right True ->
							do	writeIORef (ifTrue context) True
								return ""
						Left err ->
							do	writeIORef (ifTrue context) True
								return $ " {SSI 'elif' condition is bad: '" ++ cond ++ "' - " ++ show err ++ "} "
runDirective context "elif" argList = return $ " {SSI 'elif' directive has bad arguments: " ++ show argList ++ "} "

runDirective context "else" [] =
	do	oldIfTrue <- readIORef (ifTrue context)
		writeIORef (ifTrue context) (not oldIfTrue)
		return ""
runDirective context "else" argList = return $ " {SSI 'else' directive has bad arguments: " ++ show argList ++ "} "

runDirective context "endif" [] =
	do	oldIfTrueStack <- readIORef (ifTrueStack context)
		case oldIfTrueStack of
			newIfTrue:newStack ->
				do	writeIORef (ifTrue      context) newIfTrue
					writeIORef (ifTrueStack context) newStack
					return ""
			otherwise ->
				do	writeIORef (ifTrue context) True
					return " {SSI 'endif' is extra} "
runDirective context "endif" argList = return $ " {SSI 'endif' directive has bad arguments: " ++ show argList ++ "} "

-- Default
runDirective _ name argList = return $ " {Unknown SSI directive: " ++ name ++ " " ++ show argList ++ "} "
