module JCAB.String (
    padLeftTo,
    stripSpacesLeft, stripSpacesRight, stripSpaces,
    isPrefixOf, isSuffixOf,
    fileToUNICODE,
) where

import Data.Bits
import Data.Char
import Data.Word

-- Pads a string on the left with as many spaces as necessary
-- to bring the length of the string up to the specified number.

padLeftTo  :: Int -> String -> String
padRightTo :: Int -> String -> String

padLeftTo num string =
    let len = length string in
    if num >= len  then (replicate (num-len) ' ') ++ string
                   else string

padRightTo num string =
    let len = length string in
    string ++  if num >= len    then (replicate (num-len) ' ')
                                else ""

-- The following functions strip any existing spaces on either end of a string.

stripSpacesLeft  :: String -> String
stripSpacesRight :: String -> String
stripSpaces      :: String -> String

stripSpacesLeft [] = []
stripSpacesLeft s@(x:rest) =
    if isSpace x then
            stripSpacesLeft rest
        else
            s

stripSpacesRight [] = []
stripSpacesRight (x:rest)
    | isSpace x =
        let r = stripSpacesRight rest in
        if r == "" then
                []
            else
                x:r
    | otherwise = x:stripSpacesRight rest

stripSpaces s = stripSpacesLeft $ stripSpacesRight s

-- The following functions check whether a string's first or last few characters
-- match another string.

isPrefixOf :: String -> String -> Bool
isSuffixOf :: String -> String -> Bool

isPrefixOf p s = take (length p) s == p

isSuffixOf p s =
    if length s >= length p then
            drop (length s - length p) s == p
        else
            False

-- Stream processor to convert a string of characters that represent one byte each
-- into a string of characters, interpreting the bytes as UNICODE if necesary.
-- It uses the marker (0xFFFE or 0xFEFF) at the beginning of the string to detect
-- the UNICODE and its endianness if present.

fileToUNICODE :: String -> String

fileToUNICODE [] = []
fileToUNICODE ('\xFF':'\xFE':rest) = fileToUNICODE_ rest
    where
    fileToUNICODE_ [] = []
    fileToUNICODE_ (c0:c1:rest) = chr (ord c1 * 256 + ord c0):fileToUNICODE_ rest
fileToUNICODE ('\xFE':'\xFF':rest) = fileToUNICODE_ rest
    where
    fileToUNICODE_ [] = []
    fileToUNICODE_ (c0:c1:rest) = chr (ord c0 * 256 + ord c1):fileToUNICODE_ rest
fileToUNICODE file = file

-- Parses a string that contains UTF-8 encoded characters.

utf8ToChar :: String -> String

utf8ToChar [] = []
utf8ToChar (ch:rest) =
    let n = fromIntegral (ord ch) :: Word16 in
    if (n .&. 0xFF80) /= 0x80 then
            ch:utf8ToChar rest
        else if (n .&. 0xFFE0) == 0xC0 then
            range7FF ch ((n .&. 0x1F) `shiftL` 6) rest
        else if (n .&. 0xFFF0) == 0xE0 then
            rangeFFFF ch ((n .&. 0x0F) `shiftL` 12) rest
        else
            ch:utf8ToChar rest
    where

    range7FF ch n s@(ch2:rest) =
        let n2 = fromIntegral (ord ch2) :: Word16 in
        if (n2 .&. 0xFFC0) /= 0x80 then
                ch:utf8ToChar s
            else
                chr (fromIntegral ((n2 .&. 0x3F) + n)):utf8ToChar rest
    range7FF ch _ s = ch:utf8ToChar s

    rangeFFFF ch n s@(ch2:ch3:rest) =
        let n2 = fromIntegral (ord ch2) :: Word16 in
        let n3 = fromIntegral (ord ch3) :: Word16 in
        if (n2 .&. 0xFFC0) /= 0x80 || (n3 .&. 0xFFC0) /= 0x80 then
                ch:utf8ToChar s
            else
                chr (fromIntegral ((n3 .&. 0x3F) + ((n2 .&. 0x3F) `shiftL` 6) + n)):utf8ToChar rest
    rangeFFFF ch _ s = ch:utf8ToChar s
