module Win32.Commctl where

import Win32.System
import Win32.Window
import Win32.GDI

import Data.Bits
import Foreign.Marshal.Alloc
import Foreign.Ptr
import Foreign.Storable

newtype ICC = ICC { unICC :: DWORD } deriving (Eq, Ord, Show, Num, Bits, Storable)

iCC_LISTVIEW_CLASSES   = ICC 0x00000001 -- listview, header
iCC_TREEVIEW_CLASSES   = ICC 0x00000002 -- treeview, tooltips
iCC_BAR_CLASSES        = ICC 0x00000004 -- toolbar, statusbar, trackbar, tooltips
iCC_TAB_CLASSES        = ICC 0x00000008 -- tab, tooltips
iCC_UPDOWN_CLASS       = ICC 0x00000010 -- updown
iCC_PROGRESS_CLASS     = ICC 0x00000020 -- progress
iCC_HOTKEY_CLASS       = ICC 0x00000040 -- hotkey
iCC_ANIMATE_CLASS      = ICC 0x00000080 -- animate
iCC_WIN95_CLASSES      = ICC 0x000000FF
iCC_DATE_CLASSES       = ICC 0x00000100 -- month picker, date picker, time picker, updown
iCC_USEREX_CLASSES     = ICC 0x00000200 -- comboex
iCC_COOL_CLASSES       = ICC 0x00000400 -- rebar (coolbar) control
iCC_INTERNET_CLASSES   = ICC 0x00000800
iCC_PAGESCROLLER_CLASS = ICC 0x00001000 -- page scroller
iCC_NATIVEFNTCTL_CLASS = ICC 0x00002000 -- native font control
iCC_STANDARD_CLASSES   = ICC 0x00004000
iCC_LINK_CLASS         = ICC 0x00008000

newtype CDDS = CDDS { unCDDS :: DWORD } deriving (Eq, Ord, Show, Num, Bits, Storable)

-- drawstage flags
-- values under 0x00010000 are reserved for global custom draw values.
-- above that are for specific controls
cDDS_PREPAINT           = CDDS 0x00000001
cDDS_POSTPAINT          = CDDS 0x00000002
cDDS_PREERASE           = CDDS 0x00000003
cDDS_POSTERASE          = CDDS 0x00000004
cDDS_ITEM               = CDDS 0x00010000 -- the 0x000010000 bit means it's individual item specific
cDDS_ITEMPREPAINT       = cDDS_ITEM .|. cDDS_PREPAINT
cDDS_ITEMPOSTPAINT      = cDDS_ITEM .|. cDDS_POSTPAINT
cDDS_ITEMPREERASE       = cDDS_ITEM .|. cDDS_PREERASE
cDDS_ITEMPOSTERASE      = cDDS_ITEM .|. cDDS_POSTERASE
cDDS_SUBITEM            = CDDS 0x00020000
cDDS_SUBITEMPREPAINT    = cDDS_SUBITEM .|. cDDS_ITEMPREPAINT
cDDS_SUBITEMPOSTPAINT   = cDDS_SUBITEM .|. cDDS_ITEMPOSTPAINT
cDDS_SUBITEMPREERASE    = cDDS_SUBITEM .|. cDDS_ITEMPREERASE
cDDS_SUBITEMPOSTERASE   = cDDS_SUBITEM .|. cDDS_ITEMPOSTERASE

-- custom draw return flags
-- values under 0x00010000 are reserved for global custom draw values.
-- above that are for specific controls
cDRF_DODEFAULT          = 0x00000000 :: LRESULT
cDRF_NEWFONT            = 0x00000002 :: LRESULT
cDRF_SKIPDEFAULT        = 0x00000004 :: LRESULT
cDRF_NOTIFYPOSTPAINT    = 0x00000010 :: LRESULT
cDRF_NOTIFYITEMDRAW     = 0x00000020 :: LRESULT
cDRF_NOTIFYSUBITEMDRAW  = 0x00000020 :: LRESULT  -- flags are the same, we can distinguish by context
cDRF_NOTIFYPOSTERASE    = 0x00000040 :: LRESULT

wC_HEADER         = WndClassName "SysHeader32"       nullHINSTANCE
wC_LISTVIEW       = WndClassName "SysListView32"     nullHINSTANCE
wC_TREEVIEW       = WndClassName "SysTreeView32"     nullHINSTANCE
wC_COMBOBOXEX     = WndClassName "ComboBoxEx32"      nullHINSTANCE
wC_TABCONTROL     = WndClassName "SysTabControl32"   nullHINSTANCE
wC_IPADDRESS      = WndClassName "SysIPAddress32"    nullHINSTANCE
wC_PAGESCROLLER   = WndClassName "SysPager"          nullHINSTANCE
wC_NATIVEFONTCTLA = WndClassName "NativeFontCtl"     nullHINSTANCE
wC_BUTTON         = WndClassName "Button"            nullHINSTANCE
wC_STATIC         = WndClassName "Static"            nullHINSTANCE
wC_EDIT           = WndClassName "Edit"              nullHINSTANCE
wC_LISTBOX        = WndClassName "ListBox"           nullHINSTANCE
wC_COMBOBOX       = WndClassName "ComboBox"          nullHINSTANCE
wC_SCROLLBAR      = WndClassName "ScrollBar"         nullHINSTANCE
wC_LINK           = WndClassName "SysLink"           nullHINSTANCE
wC_TOOLTIPS       = WndClassName "tooltips_class32"  nullHINSTANCE
wC_TRACKBAR       = WndClassName "msctls_trackbar32" nullHINSTANCE
wC_UPDOWN         = WndClassName "msctls_updown32"   nullHINSTANCE
wC_PROGRESS       = WndClassName "msctls_progress32" nullHINSTANCE
wC_HOTKEY         = WndClassName "msctls_hotkey32"   nullHINSTANCE
wC_ANIMATE        = WndClassName "SysAnimate32"      nullHINSTANCE
wC_MONTHCAL       = WndClassName "SysMonthCal32"     nullHINSTANCE
wC_DATETIMEPICK   = WndClassName "SysDateTimePick32" nullHINSTANCE

initCommonControls :: ICC -> IO BOOL

initCommonControls (ICC icc) =
    allocaBytes 8 $ \ptr -> do
        pokeByteOff ptr 0 (8 :: DWORD)
        pokeByteOff ptr 4 icc
        c_InitCommonControlsEx $ ptr

foreign import stdcall unsafe "commctrl.h InitCommonControlsEx" c_InitCommonControlsEx :: PVOID -> IO BOOL



cOMCTL32_VERSION = 6

lVM_FIRST               = WMSG 0x1000      -- ListView messages
tV_FIRST                = WMSG 0x1100      -- TreeView messages
hDM_FIRST               = WMSG 0x1200      -- Header messages
tCM_FIRST               = WMSG 0x1300      -- Tab control messages

pGM_FIRST               = WMSG 0x1400      -- Pager control messages
eCM_FIRST               = WMSG 0x1500      -- Edit control messages
bCM_FIRST               = WMSG 0x1600      -- Button control messages
cBM_FIRST               = WMSG 0x1700      -- Combobox control messages
cCM_FIRST               = WMSG 0x2000      -- Common control shared messages
cCM_LAST                = cCM_FIRST + WMSG 0x200

nM_FIRST                = WNOTIFICATION (-  0)       -- generic to all controls
nM_LAST                 = WNOTIFICATION (- 99)

lVN_FIRST               = WNOTIFICATION (-100)       -- listview
lVN_LAST                = WNOTIFICATION (-199)

-- Property sheet reserved      -200 -  -299 - see prsht.h

hDN_FIRST               = WNOTIFICATION (-300)       -- header
hDN_LAST                = WNOTIFICATION (-399)

tVN_FIRST               = WNOTIFICATION (-400)       -- treeview
tVN_LAST                = WNOTIFICATION (-499)

tTN_FIRST               = WNOTIFICATION (-520)       -- tooltips
tTN_LAST                = WNOTIFICATION (-549)

tCN_FIRST               = WNOTIFICATION (-550)       -- tab control
tCN_LAST                = WNOTIFICATION (-580)

-- Shell reserved               = WNOTIFICATION (-580) -  = WNOTIFICATION (-589)

cDN_FIRST               = WNOTIFICATION (-601)       -- common dialog = WNOTIFICATION (new)
cDN_LAST                = WNOTIFICATION (-699)

tBN_FIRST               = WNOTIFICATION (-700)       -- toolbar
tBN_LAST                = WNOTIFICATION (-720)

uDN_FIRST               = WNOTIFICATION (-721)        -- updown
uDN_LAST                = WNOTIFICATION (-740)

mCN_FIRST               = WNOTIFICATION (-750)       -- monthcal
mCN_LAST                = WNOTIFICATION (-759)

dTN_FIRST               = WNOTIFICATION (-760)       -- datetimepick
dTN_LAST                = WNOTIFICATION (-799)

cBEN_FIRST              = WNOTIFICATION (-800)       -- combo box ex
cBEN_LAST               = WNOTIFICATION (-830)

rBN_FIRST               = WNOTIFICATION (-831)       -- rebar
rBN_LAST                = WNOTIFICATION (-859)

iPN_FIRST               = WNOTIFICATION (-860)       -- internet address
iPN_LAST                = WNOTIFICATION (-879)       -- internet address

sBN_FIRST               = WNOTIFICATION (-880)       -- status bar
sBN_LAST                = WNOTIFICATION (-899)

pGN_FIRST               = WNOTIFICATION (-900)       -- Pager Control
pGN_LAST                = WNOTIFICATION (-950)

wMN_FIRST               = WNOTIFICATION (-1000)
wMN_LAST                = WNOTIFICATION (-1200)

bCN_FIRST               = WNOTIFICATION (-1250)
bCN_LAST                = WNOTIFICATION (-1350)



cCM_SETBKCOLOR          = cCM_FIRST + WMSG 1 -- lParam is bkColor

{-
typedef struct tagCOLORSCHEME {
   DWORD            dwSize;
   COLORREF         clrBtnHighlight;       // highlight color
   COLORREF         clrBtnShadow;          // shadow color
} COLORSCHEME, *LPCOLORSCHEME;
-}

cCM_SETCOLORSCHEME      = cCM_FIRST + WMSG 2 -- lParam is color scheme
cCM_GETCOLORSCHEME      = cCM_FIRST + WMSG 3 -- fills in COLORSCHEME pointed to by lParam
cCM_GETDROPTARGET       = cCM_FIRST + WMSG 4
cCM_SETUNICODEFORMAT    = cCM_FIRST + WMSG 5
cCM_GETUNICODEFORMAT    = cCM_FIRST + WMSG 6

cCM_SETVERSION          = cCM_FIRST + WMSG 0x7
cCM_GETVERSION          = cCM_FIRST + WMSG 0x8
cCM_SETNOTIFYWINDOW     = cCM_FIRST + WMSG 0x9 -- wParam == hwndParent.
cCM_SETWINDOWTHEME      = cCM_FIRST + WMSG 0xb
cCM_DPISCALE            = cCM_FIRST + WMSG 0xc -- wParam == Awareness

nM_OUTOFMEMORY          = nM_FIRST - WNOTIFICATION  1
nM_CLICK                = nM_FIRST - WNOTIFICATION  2   -- uses NMCLICK struct
nM_DBLCLK               = nM_FIRST - WNOTIFICATION  3
nM_RETURN               = nM_FIRST - WNOTIFICATION  4
nM_RCLICK               = nM_FIRST - WNOTIFICATION  5   -- uses NMCLICK struct
nM_RDBLCLK              = nM_FIRST - WNOTIFICATION  6
nM_SETFOCUS             = nM_FIRST - WNOTIFICATION  7
nM_KILLFOCUS            = nM_FIRST - WNOTIFICATION  8
nM_CUSTOMDRAW           = nM_FIRST - WNOTIFICATION 12
nM_HOVER                = nM_FIRST - WNOTIFICATION 13
nM_NCHITTEST            = nM_FIRST - WNOTIFICATION 14   -- uses NMMOUSE struct
nM_KEYDOWN              = nM_FIRST - WNOTIFICATION 15   -- uses NMKEY struct
nM_RELEASEDCAPTURE      = nM_FIRST - WNOTIFICATION 16
nM_SETCURSOR            = nM_FIRST - WNOTIFICATION 17   -- uses NMMOUSE struct
nM_CHAR                 = nM_FIRST - WNOTIFICATION 18   -- uses NMCHAR struct
nM_TOOLTIPSCREATED      = nM_FIRST - WNOTIFICATION 19   -- notify of when the tooltips window is create
nM_LDOWN                = nM_FIRST - WNOTIFICATION 20
nM_RDOWN                = nM_FIRST - WNOTIFICATION 21
nM_THEMECHANGED         = nM_FIRST - WNOTIFICATION 22




eS_LEFT             = WSTYLE 0x0000
eS_CENTER           = WSTYLE 0x0001
eS_RIGHT            = WSTYLE 0x0002
eS_MULTILINE        = WSTYLE 0x0004
eS_UPPERCASE        = WSTYLE 0x0008
eS_LOWERCASE        = WSTYLE 0x0010
eS_PASSWORD         = WSTYLE 0x0020
eS_AUTOVSCROLL      = WSTYLE 0x0040
eS_AUTOHSCROLL      = WSTYLE 0x0080
eS_NOHIDESEL        = WSTYLE 0x0100
eS_OEMCONVERT       = WSTYLE 0x0400
eS_READONLY         = WSTYLE 0x0800
eS_WANTRETURN       = WSTYLE 0x1000
eS_NUMBER           = WSTYLE 0x2000





lVS_ICON                = WSTYLE 0x0000
lVS_REPORT              = WSTYLE 0x0001
lVS_SMALLICON           = WSTYLE 0x0002
lVS_LIST                = WSTYLE 0x0003
lVS_TYPEMASK            = WSTYLE 0x0003
lVS_SINGLESEL           = WSTYLE 0x0004
lVS_SHOWSELALWAYS       = WSTYLE 0x0008
lVS_SORTASCENDING       = WSTYLE 0x0010
lVS_SORTDESCENDING      = WSTYLE 0x0020
lVS_SHAREIMAGELISTS     = WSTYLE 0x0040
lVS_NOLABELWRAP         = WSTYLE 0x0080
lVS_AUTOARRANGE         = WSTYLE 0x0100
lVS_EDITLABELS          = WSTYLE 0x0200
lVS_OWNERDATA           = WSTYLE 0x1000

lVS_NOSCROLL            = WSTYLE 0x2000

lVS_TYPESTYLEMASK       = WSTYLE 0xfc00

lVS_ALIGNTOP            = WSTYLE 0x0000
lVS_ALIGNLEFT           = WSTYLE 0x0800
lVS_ALIGNMASK           = WSTYLE 0x0c00

lVS_OWNERDRAWFIXED      = WSTYLE 0x0400
lVS_NOCOLUMNHEADER      = WSTYLE 0x4000
lVS_NOSORTHEADER        = WSTYLE 0x8000

newtype LVEXSTYLE = LVEXSTYLE { unLVEXSTYLE :: DWORD } deriving (Eq, Ord, Show, Num, Enum, Real, Integral, Bits, Storable)

lVS_EX_GRIDLINES        = LVEXSTYLE 0x00000001
lVS_EX_SUBITEMIMAGES    = LVEXSTYLE 0x00000002
lVS_EX_CHECKBOXES       = LVEXSTYLE 0x00000004
lVS_EX_TRACKSELECT      = LVEXSTYLE 0x00000008
lVS_EX_HEADERDRAGDROP   = LVEXSTYLE 0x00000010
lVS_EX_FULLROWSELECT    = LVEXSTYLE 0x00000020 -- applies to report mode only
lVS_EX_ONECLICKACTIVATE = LVEXSTYLE 0x00000040
lVS_EX_TWOCLICKACTIVATE = LVEXSTYLE 0x00000080
lVS_EX_FLATSB           = LVEXSTYLE 0x00000100
lVS_EX_REGIONAL         = LVEXSTYLE 0x00000200
lVS_EX_INFOTIP          = LVEXSTYLE 0x00000400 -- listview does InfoTips for you
lVS_EX_UNDERLINEHOT     = LVEXSTYLE 0x00000800
lVS_EX_UNDERLINECOLD    = LVEXSTYLE 0x00001000
lVS_EX_MULTIWORKAREAS   = LVEXSTYLE 0x00002000
lVS_EX_LABELTIP         = LVEXSTYLE 0x00004000 -- listview unfolds partly hidden labels if it does not have infotip text
lVS_EX_BORDERSELECT     = LVEXSTYLE 0x00008000 -- border selection style instead of highlight
lVS_EX_DOUBLEBUFFER     = LVEXSTYLE 0x00010000
lVS_EX_HIDELABELS       = LVEXSTYLE 0x00020000
lVS_EX_SINGLEROW        = LVEXSTYLE 0x00040000
lVS_EX_SNAPTOGRID       = LVEXSTYLE 0x00080000  -- Icons automatically snap to grid.
lVS_EX_SIMPLESELECT     = LVEXSTYLE 0x00100000  -- Also changes overlay rendering to top right for icon mode.

lVM_SETITEMCOUNT             = lVM_FIRST + WMSG 47
lVM_SETEXTENDEDLISTVIEWSTYLE = lVM_FIRST + WMSG 54
lVM_SETITEM                  = lVM_FIRST + WMSG 76
lVM_INSERTITEM               = lVM_FIRST + WMSG 77
lVM_INSERTCOLUMN             = lVM_FIRST + WMSG 97

lVN_ITEMCHANGING        = lVN_FIRST - WNOTIFICATION  0
lVN_ITEMCHANGED         = lVN_FIRST - WNOTIFICATION  1
lVN_INSERTITEM          = lVN_FIRST - WNOTIFICATION  2
lVN_DELETEITEM          = lVN_FIRST - WNOTIFICATION  3
lVN_DELETEALLITEMS      = lVN_FIRST - WNOTIFICATION  4
lVN_BEGINLABELEDIT      = lVN_FIRST - WNOTIFICATION 75
lVN_ENDLABELEDITA       = lVN_FIRST - WNOTIFICATION  6
lVN_ENDLABELEDITW       = lVN_FIRST - WNOTIFICATION 76
lVN_COLUMNCLICK         = lVN_FIRST - WNOTIFICATION  8
lVN_BEGINDRAG           = lVN_FIRST - WNOTIFICATION  9
lVN_BEGINRDRAG          = lVN_FIRST - WNOTIFICATION 11

lVN_ODCACHEHINT         = lVN_FIRST - WNOTIFICATION 13
lVN_ODFINDITEM          = lVN_FIRST - WNOTIFICATION 79

lVN_ITEMACTIVATE        = lVN_FIRST - WNOTIFICATION 14
lVN_ODSTATECHANGED      = lVN_FIRST - WNOTIFICATION 15

lVN_HOTTRACK            = lVN_FIRST - WNOTIFICATION 21

lVN_GETDISPINFO         = lVN_FIRST - WNOTIFICATION 77
lVN_SETDISPINFO         = lVN_FIRST - WNOTIFICATION 78

listView_SetItemCount               :: HWND -> INT -> IO ()
listView_SetItemCountEx             :: HWND -> INT -> INT -> IO ()
listView_SetExtendedListViewStyle   :: HWND -> LVEXSTYLE -> IO ()
listView_SetExtendedListViewStyleEx :: HWND -> LVEXSTYLE -> LVEXSTYLE -> IO ()
listView_SetItemCText               :: HWND -> INT -> INT -> PSTR -> IO INT
listView_SetItemText                :: HWND -> INT -> INT -> String -> IO INT
listView_InsertItemCText            :: HWND -> INT -> INT -> PSTR -> IO INT
listView_InsertItemText             :: HWND -> INT -> INT -> String -> IO INT
listView_SetInsertColumnWidthText   :: HWND -> INT -> INT -> String -> IO INT


newtype LVIF = LVIF { unLVIF :: DWORD } deriving (Eq, Ord, Show, Num, Enum, Real, Integral, Bits, Storable)

lVIF_TEXT               = LVIF 0x0001
lVIF_IMAGE              = LVIF 0x0002
lVIF_PARAM              = LVIF 0x0004
lVIF_STATE              = LVIF 0x0008
lVIF_INDENT             = LVIF 0x0010
lVIF_NORECOMPUTE        = LVIF 0x0800
lVIF_GROUPID            = LVIF 0x0100
lVIF_COLUMNS            = LVIF 0x0200

newtype LVCF = LVCF { unLVCF :: DWORD } deriving (Eq, Ord, Show, Num, Enum, Real, Integral, Bits, Storable)

lVCF_FMT                = LVCF 0x0001
lVCF_WIDTH              = LVCF 0x0002
lVCF_TEXT               = LVCF 0x0004
lVCF_SUBITEM            = LVCF 0x0008
lVCF_IMAGE              = LVCF 0x0010
lVCF_ORDER              = LVCF 0x0020

listView_SetItemCount hwnd count = do
    sendMessage hwnd lVM_SETITEMCOUNT (fromIntegral count) 0
    return ()

listView_SetItemCountEx hwnd count flags = do
    sendMessage hwnd lVM_SETITEMCOUNT (fromIntegral count) (fromIntegral flags)
    return ()

listView_SetExtendedListViewStyleEx hwnd mask style = do
    sendMessage hwnd lVM_SETEXTENDEDLISTVIEWSTYLE (fromIntegral mask) (fromIntegral style)
    return ()

listView_SetExtendedListViewStyle hwnd style = listView_SetExtendedListViewStyleEx hwnd (LVEXSTYLE 0) style

listView_SetItemCText hwnd item col text =
    allocaBytes 52 $ \c_item -> do
        pokeByteOff c_item  0 (lVIF_TEXT)
        pokeByteOff c_item  4 item
        pokeByteOff c_item  8 col
        pokeByteOff c_item  4 item
        pokeByteOff c_item 20 text
        result <- sendMessage hwnd lVM_SETITEM 0 (fromIntegral (c_item `minusPtr` nullPtr))
        return (fromIntegral result)

listView_SetItemText hwnd item col text =
    withPSTR text $ \c_text -> listView_SetItemCText hwnd item col c_text

listView_InsertItemCText hwnd item col text =
    allocaBytes 52 $ \c_item -> do
        pokeByteOff c_item  0 (lVIF_TEXT)
        pokeByteOff c_item  4 item
        pokeByteOff c_item  8 col
        pokeByteOff c_item  4 item
        pokeByteOff c_item 20 text
        result <- sendMessage hwnd lVM_INSERTITEM 0 (fromIntegral (c_item `minusPtr` nullPtr))
        return (fromIntegral result)

listView_InsertItemText hwnd item col text =
    withPSTR text $ \c_text -> listView_InsertItemCText hwnd item col c_text

listView_SetInsertColumnWidthText hwnd col width text =
    allocaBytes 32 $ \c_column ->
    withPSTR text $ \c_text -> do
        pokeByteOff c_column  0 (lVCF_WIDTH .|. lVCF_TEXT)
        pokeByteOff c_column  8 width
        pokeByteOff c_column 12 c_text
        result <- sendMessage hwnd lVM_INSERTCOLUMN (fromIntegral col) (fromIntegral (c_column `minusPtr` nullPtr))
        return (fromIntegral result)




-- Window message structures

data LVITEM         ; type PLVITEM         = Ptr LVITEM
data NMLVDISPINFO   ; type PNMLVDISPINFO   = Ptr NMLVDISPINFO
data NMLVCUSTOMDRAW ; type PNMLVCUSTOMDRAW = Ptr NMLVCUSTOMDRAW
data NMCUSTOMDRAW   ; type PNMCUSTOMDRAW   = Ptr NMCUSTOMDRAW

-- Window message structure hierarchy

instance ParentCast NMLVDISPINFO NMHDR
instance ParentCast NMLVDISPINFO LVITEM where upCast = castPtrOffset 12

instance ParentCast NMCUSTOMDRAW NMHDR

instance ParentCast NMLVCUSTOMDRAW NMHDR
instance ParentCast NMLVCUSTOMDRAW NMCUSTOMDRAW

-- Window message structure fields

instance Storable LVITEM where { sizeOf m = 52; alignment m = 4; peek = undefined; poke = undefined }

pLVITEMmask       = (undefined, undefined,  0) :: StructField LVITEM LVIF
pLVITEMiItem      = (undefined, undefined,  4) :: StructField LVITEM INT
pLVITEMiSubItem   = (undefined, undefined,  8) :: StructField LVITEM INT
pLVITEMstate      = (undefined, undefined, 12) :: StructField LVITEM UINT
pLVITEMstateMask  = (undefined, undefined, 16) :: StructField LVITEM UINT
pLVITEMpszText    = (undefined, undefined, 20) :: StructField LVITEM PSTR
pLVITEMcchTextMax = (undefined, undefined, 24) :: StructField LVITEM INT
pLVITEMiImage     = (undefined, undefined, 28) :: StructField LVITEM INT
pLVITEMlParam     = (undefined, undefined, 32) :: StructField LVITEM LPARAM
pLVITEMiIndent    = (undefined, undefined, 36) :: StructField LVITEM INT
pLVITEMiGroupId   = (undefined, undefined, 40) :: StructField LVITEM INT
pLVITEMcColumns   = (undefined, undefined, 44) :: StructField LVITEM UINT
pLVITEMpuColumns  = (undefined, undefined, 48) :: StructField LVITEM PUINT

instance Storable NMCUSTOMDRAW where { sizeOf m = 48; alignment m = 4; peek = undefined; poke = undefined }

pNMCUSTOMDRAWdwDrawStage = (undefined, undefined, 12) :: StructField NMCUSTOMDRAW CDDS
pNMCUSTOMDRAWhdc         = (undefined, undefined, 16) :: StructField NMCUSTOMDRAW HDC
pNMCUSTOMDRAWrc          = (undefined, undefined, 20) :: StructField NMCUSTOMDRAW RECT
pNMCUSTOMDRAWdwItemSpec  = (undefined, undefined, 36) :: StructField NMCUSTOMDRAW DWORD
pNMCUSTOMDRAWuItemState  = (undefined, undefined, 40) :: StructField NMCUSTOMDRAW UINT
pNMCUSTOMDRAWlItemlParam = (undefined, undefined, 44) :: StructField NMCUSTOMDRAW LPARAM

instance Storable NMLVCUSTOMDRAW where { sizeOf m = 104; alignment m = 4; peek = undefined; poke = undefined }

pNMLVCUSTOMDRAWclrText     = (undefined, undefined,  48) :: StructField NMLVCUSTOMDRAW COLORREF
pNMLVCUSTOMDRAWclrTextBk   = (undefined, undefined,  52) :: StructField NMLVCUSTOMDRAW COLORREF
pNMLVCUSTOMDRAWiSubItem    = (undefined, undefined,  56) :: StructField NMLVCUSTOMDRAW INT
pNMLVCUSTOMDRAWdwItemType  = (undefined, undefined,  60) :: StructField NMLVCUSTOMDRAW DWORD
pNMLVCUSTOMDRAWclrFace     = (undefined, undefined,  64) :: StructField NMLVCUSTOMDRAW COLORREF
pNMLVCUSTOMDRAWiIconEffect = (undefined, undefined,  68) :: StructField NMLVCUSTOMDRAW INT
pNMLVCUSTOMDRAWiIconPhase  = (undefined, undefined,  72) :: StructField NMLVCUSTOMDRAW INT
pNMLVCUSTOMDRAWiPartId     = (undefined, undefined,  76) :: StructField NMLVCUSTOMDRAW INT
pNMLVCUSTOMDRAWiStateId    = (undefined, undefined,  80) :: StructField NMLVCUSTOMDRAW INT
pNMLVCUSTOMDRAWrcText      = (undefined, undefined,  84) :: StructField NMLVCUSTOMDRAW RECT
pNMLVCUSTOMDRAWuAlign      = (undefined, undefined, 100) :: StructField NMLVCUSTOMDRAW UINT
