module Win32.GDI.API where

import Win32.GDI.Types
import Win32.GDI.Constants
import Win32.GDI.Foreign

import Win32.System

import Data.Bits
import Foreign.Marshal.Alloc
import Foreign.Marshal.Array
import Foreign.Marshal.Utils
import Foreign.Ptr
import Foreign.Storable
import System.IO.Unsafe




-- Handles

castToHBRUSH   :: Integral a => a -> HBRUSH
castToHPEN     :: Integral a => a -> HPEN
castToHFONT    :: Integral a => a -> HFONT
castToHDC      :: Integral a => a -> HDC

castFromHBRUSH :: Integral a => HBRUSH -> a
castFromHPEN   :: Integral a => HPEN   -> a
castFromHFONT  :: Integral a => HFONT  -> a
castFromHDC    :: Integral a => HDC    -> a

nullHBRUSH     :: HBRUSH
nullHPEN       :: HPEN
nullHFONT      :: HFONT
nullHDC        :: HDC


castToHBRUSH l = HBRUSH    (castToPtr l)
castToHPEN   l = HPEN      (castToPtr l)
castToHFONT  l = HFONT     (castToPtr l)
castToHDC    l = HDC       (castToPtr l)

castFromHBRUSH (HBRUSH p) = castFromPtr p
castFromHPEN   (HPEN   p) = castFromPtr p
castFromHFONT  (HFONT  p) = castFromPtr p
castFromHDC    (HDC    p) = castFromPtr p

nullHBRUSH = HBRUSH nullPtr
nullHPEN   = HPEN   nullPtr
nullHFONT  = HFONT  nullPtr
nullHDC    = HDC    nullPtr




-- Colors

rgb :: (Integral a, Integral b, Integral c) => a -> b -> c -> COLORREF

getRValue :: Integral a => COLORREF -> a
getGValue :: Integral a => COLORREF -> a
getBValue :: Integral a => COLORREF -> a


rgb r g b = COLORREF $ (fromIntegral r) + ((fromIntegral g) `shiftL` 8) + ((fromIntegral b) `shiftL` 16)

getRValue (COLORREF color) = fromIntegral ( color              .&. 0xFF)
getGValue (COLORREF color) = fromIntegral ((color `shiftR`  8) .&. 0xFF)
getBValue (COLORREF color) = fromIntegral ((color `shiftR` 16) .&. 0xFF)




-- Brushes

selectBrush     :: HDC -> HBRUSH -> IO HBRUSH
setDCBrushColor :: HDC -> COLORREF -> IO COLORREF
withSelectBrush :: HDC -> HBRUSH -> IO a -> IO (Maybe a)


selectBrush     = c_SelectBrush
setDCBrushColor = c_SetDCBrushColor

withSelectBrush dc brush a = do
    prev <- selectBrush dc brush
    if prev /= nullHBRUSH then do
            result <- a
            selectBrush dc prev
            return $ Just result
        else
            return Nothing




-- Pens

createPen     :: PENSTYLE -> INT -> COLORREF -> IO HPEN
deletePen     :: HPEN -> IO BOOL
withCreatePen :: PENSTYLE -> INT -> COLORREF -> (HPEN -> IO a) -> IO a

selectPen     :: HDC -> HPEN   -> IO HPEN
setDCPenColor :: HDC -> COLORREF -> IO COLORREF
withSelectPen :: HDC -> HPEN   -> IO a -> IO (Maybe a)


createPen = c_CreatePen
deletePen = c_DeletePen

withCreatePen s w c f = do
    pen <- createPen s w c
    result <- f pen
    deletePen pen
    return result

selectPen     = c_SelectPen
setDCPenColor = c_SetDCPenColor

withSelectPen dc pen a = do
    prev <- selectPen dc pen
    if prev /= nullHPEN then do
            result <- a
            selectPen dc prev
            return $ Just result
        else
            return Nothing




-- Fonts

selectFont     :: HDC -> HFONT  -> IO HFONT
withSelectFont :: HDC -> HFONT  -> IO a -> IO (Maybe a)


selectFont  = c_SelectFont

withSelectFont dc font a = do
    prev <- selectFont dc font
    if prev /= nullHFONT then do
            result <- a
            selectFont dc prev
            return $ Just result
        else
            return Nothing




-- Filled shapes

chord       :: HDC -> RECT -> POINT -> POINT -> IO BOOL
pie         :: HDC -> RECT -> POINT -> POINT -> IO BOOL
ellipse     :: HDC -> RECT -> IO BOOL
rectangle   :: HDC -> RECT -> IO BOOL
roundRect   :: HDC -> RECT -> INT -> INT -> IO BOOL
fillRect    :: HDC -> RECT -> HBRUSH -> IO BOOL
frameRect   :: HDC -> RECT -> HBRUSH -> IO BOOL
invertRect  :: HDC -> RECT -> IO BOOL
polygon     :: HDC -> [POINT] -> IO BOOL
polyPolygon :: HDC -> [[POINT]] -> IO BOOL

chord       dc (RECT l t r b) (POINT fx fy) (POINT sx sy) = c_Chord     dc l t r b fx fy sx sy
pie         dc (RECT l t r b) (POINT fx fy) (POINT sx sy) = c_Chord     dc l t r b fx fy sx sy
ellipse     dc (RECT l t r b)                             = c_Ellipse   dc l t r b
rectangle   dc (RECT l t r b)                             = c_Rectangle dc l t r b
roundRect   dc (RECT l t r b) ew eh                       = c_RoundRect dc l t r b ew eh
fillRect    dc rect brush                                 = with rect           $ \prect         -> c_FillRect   dc prect brush
frameRect   dc rect brush                                 = with rect           $ \prect         -> c_FrameRect  dc prect brush
invertRect  dc rect                                       = with rect           $ \prect         -> c_InvertRect dc prect
polygon     dc points                                     = withArrayLen points $ \count ppoints -> c_Polygon    dc ppoints (fromIntegral count)

polyPolygon dc polygons =
    withArrayLen ((map (fromIntegral . length) polygons) :: [INT]) $ \count ppolycounts ->
    withArray (concat polygons) $ \ppoints ->
    c_PolyPolygon dc ppoints ppolycounts (fromIntegral count)




-- Text output

textOut              :: HDC -> POINT -> String -> IO BOOL
getTextExtentPoint32 :: HDC -> String -> IO SIZE


textOut dc (POINT x y) text =
    withPSTRLen text $ \(c_Text, c_Len) -> do
    c_TextOut dc x y c_Text (fromIntegral c_Len)

getTextExtentPoint32 dc text =
    alloca $ \psize ->
    withPSTRLen text $ \(c_Text, c_Len) -> do
        result <- c_GetTextExtentPoint32 dc c_Text (fromIntegral c_Len) psize
        if result then peek psize
                  else return $ SIZE 0 0

