module Win32.GDI.Types where

import Win32.System.Types

import Data.Bits
import Data.Typeable
import Foreign.Marshal.Alloc
import Foreign.Ptr
import Foreign.Storable

-- Handles

newtype HBRUSH = HBRUSH PVOID deriving (Eq, Show, Storable)
newtype HPEN   = HPEN   PVOID deriving (Eq, Show, Storable)
newtype HFONT  = HFONT  PVOID deriving (Eq, Show, Storable)
newtype HDC    = HDC    PVOID deriving (Eq, Show, Storable)

unHBRUSH (HBRUSH p) = p
unHPEN   (HPEN   p) = p
unHFONT  (HFONT  p) = p
unHDC    (HDC    p) = p

-- Enumeration types

newtype COLORTYPE    = COLORTYPE    UINT  deriving (Eq, Ord, Show, Enum, Num, Real, Integral, Bits, Storable, Typeable)
newtype COLORREF     = COLORREF     DWORD deriving (Eq, Ord, Show, Enum, Num, Real, Integral, Bits, Storable, Typeable)
newtype STOCK_BRUSH  = STOCK_BRUSH  INT   deriving (Eq, Ord, Show, Enum, Num, Real, Integral, Bits, Storable, Typeable)
newtype STOCK_PEN    = STOCK_PEN    INT   deriving (Eq, Ord, Show, Enum, Num, Real, Integral, Bits, Storable, Typeable)
newtype PENSTYLE     = PENSTYLE     ULONG deriving (Eq, Ord, Show, Enum, Num, Real, Integral, Bits, Storable, Typeable)
newtype STOCK_FONT   = STOCK_FONT   INT   deriving (Eq, Ord, Show, Enum, Num, Real, Integral, Bits, Storable, Typeable)

unCOLORTYPE   (COLORTYPE   v) = v
unCOLORREF    (COLORREF    v) = v
unSTOCK_BRUSH (STOCK_BRUSH v) = v
unSTOCK_PEN   (STOCK_PEN   v) = v
unPENSTYLE    (PENSTYLE    v) = v
unSTOCK_FONT  (STOCK_FONT  v) = v

-- Geometry types

data POINT = POINT LONG LONG
data SIZE  = SIZE  LONG LONG
data RECT  = RECT  LONG LONG LONG LONG

type PPOINT = Ptr POINT
type PSIZE  = Ptr SIZE
type PRECT  = Ptr RECT

sizeOfLONG = sizeOf (undefined :: LONG)

getPOINTX (POINT x _) = x
getPOINTY (POINT _ y) = y

instance Eq POINT where
    (POINT x1 y1) == (POINT x2 y2) = x1 == x2 && y1 == y2

instance Storable POINT where
    sizeOf _           = sizeOfLONG * 2
    alignment _        = alignment (undefined :: LONG)
    peek p             = do x <- peekByteOff p 0
                            y <- peekByteOff p (sizeOfLONG)
                            return $ POINT x y
    poke p (POINT x y) = do pokeByteOff p 0            x
                            pokeByteOff p (sizeOfLONG) y

getSIZEX (SIZE cx _) = cx
getSIZEY (SIZE _ cy) = cy

instance Eq SIZE where
    (SIZE cx1 cy1) == (SIZE cx2 cy2) = cx1 == cx2 && cy1 == cy2

instance Storable SIZE where
    sizeOf _            = sizeOfLONG * 2
    alignment _         = alignment (undefined :: LONG)
    peek p              = do cx <- peekByteOff p 0
                             cy <- peekByteOff p (sizeOfLONG)
                             return $ SIZE cx cy
    poke p (SIZE cx cy) = do pokeByteOff p 0            cx
                             pokeByteOff p (sizeOfLONG) cy

getRECTLeft   (RECT l _ _ _) = l
getRECTTop    (RECT _ t _ _) = t
getRECTRight  (RECT _ _ r _) = r
getRECTBottom (RECT _ _ _ b) = b
getRECTWidth  (RECT l _ r _) = r-l
getRECTHeight (RECT _ t _ b) = b-t

getRECTLeftTop     (RECT l t _ _) = POINT l t
getRECTRightTop    (RECT _ t r _) = POINT r t
getRECTLeftBottom  (RECT l _ _ b) = POINT l b
getRECTRightBottom (RECT _ _ r b) = POINT r b

getRECTSize (RECT l t r b) = SIZE (r-l) (b-t)

setRECTLeft   (RECT _ t r b) l = RECT l t r b
setRECTTop    (RECT l _ r b) t = RECT l t r b
setRECTRight  (RECT l t _ b) r = RECT l t r b
setRECTBottom (RECT l t r _) b = RECT l t r b
setRECTWidth  (RECT l t _ b) w = RECT l t (l+w) b
setRECTHeight (RECT l t r _) h = RECT l t r (t+h)

setRECTLeftTop     (RECT _ _ r b) (POINT l t) = RECT l t r b
setRECTRightTop    (RECT l _ _ b) (POINT r t) = RECT l t r b
setRECTLeftBottom  (RECT _ t r _) (POINT l b) = RECT l t r b
setRECTRightBottom (RECT l t _ _) (POINT r b) = RECT l t r b

setRECTSize   (RECT l t _ _) (SIZE w h) = RECT l t (l+w) (t+h)

makeRECTPointSize (POINT x y) (SIZE cx cy) = RECT x y (x+cx) (y+cy)
makeRECTxywh x y cx cy                     = RECT x y (x+cx) (y+cy)

makeRECTSize (SIZE cx cy) = RECT 0 0 cx cy
makeRECTwh cx cy          = RECT 0 0 cx cy

shrinkRECT sl st sr sb (RECT l t r b) = RECT (l+sl) (t+st) (r-sr) (b-sb)
expandRECT sl st sr sb (RECT l t r b) = RECT (l-sl) (t-st) (r+sr) (b+sb)

moveRECT (SIZE cx cy) (RECT l t r b) = RECT (l+cx) (t+cy) (r+cx) (b+cy)
moveRECTxy cx cy (RECT l t r b)      = RECT (l+cx) (t+cy) (r+cx) (b+cy)

cW_USEDEFAULT = 0x80000000 :: LONG

pointDefault = POINT cW_USEDEFAULT cW_USEDEFAULT
sizeDefault  = SIZE  cW_USEDEFAULT cW_USEDEFAULT
rectDefault  = makeRECTxywh cW_USEDEFAULT cW_USEDEFAULT cW_USEDEFAULT cW_USEDEFAULT
nullRECT     = RECT 0 0 0 0

instance Eq RECT where
    (RECT l1 t1 r1 b1) == (RECT l2 t2 r2 b2) = l1 == l2 && t1 == t2 && r1 == r2 && b1 == b2

instance Storable RECT where
    sizeOf _              = sizeOfLONG * 4
    alignment _           = alignment (undefined :: LONG)
    peek p                = do l <- peekByteOff p 0
                               t <- peekByteOff p (sizeOfLONG    )
                               r <- peekByteOff p (sizeOfLONG * 2)
                               b <- peekByteOff p (sizeOfLONG * 3)
                               return $ RECT l t r b

    poke p (RECT l t r b) = do pokeByteOff p 0                l
                               pokeByteOff p (sizeOfLONG    ) t
                               pokeByteOff p (sizeOfLONG * 2) r
                               pokeByteOff p (sizeOfLONG * 3) b

-- Paint struct

data PAINTSTRUCT = PAINTSTRUCT {
    psHDC       :: HDC,
    psErase     :: BOOL,
    psRect      :: RECT,
    psRestore   :: BOOL,
    psIncUpdate :: BOOL
}

type PPAINTSTRUCT = Ptr PAINTSTRUCT

instance Storable PAINTSTRUCT where
    sizeOf    m = 60
    alignment m =  4
    peek p      = do hdc       <- peekByteOff p  0
                     erase     <- peekByteOff p  4
                     rect      <- peekByteOff p  8
                     restore   <- peekByteOff p 24
                     incUpdate <- peekByteOff p 25
                     return $ PAINTSTRUCT hdc erase rect restore incUpdate
    poke p    m = do pokeByteOff p  0 $ psHDC       m
                     pokeByteOff p  4 $ psErase     m
                     pokeByteOff p  8 $ psRect      m
                     pokeByteOff p 24 $ psRestore   m
                     pokeByteOff p 25 $ psIncUpdate m

sizeOfPAINTSTRUCT = sizeOf (undefined :: PAINTSTRUCT)
withPAINTSTRUCT   = allocaBytes sizeOfPAINTSTRUCT

