module Win32.System.API where

import Win32.System.Types
import Win32.System.Constants
import Win32.System.Foreign

import Control.Monad
import Data.Bits
import Foreign.C.String
import Foreign.Marshal.Alloc
import Foreign.Marshal.Array
import Foreign.Ptr
import Foreign.Storable
import System.IO
import System.IO.Unsafe




-- Initialization

initGUI :: IO ()


initGUI = catch (putStr " \b" >> hFlush stdout) $ \_ -> do
    fd <- open "nul" 2 0
    dup2 fd 0
    dup2 fd 1
    dup2 fd 2
    return ()

-- Primitive types

loWORD :: (Bits a, Integral a, Integral b) => a -> b
hiWORD :: (Bits a, Integral a, Integral b) => a -> b


loWORD p = fromIntegral ( p              .&. 0xFFFF)
hiWORD p = fromIntegral ((p `shiftR` 16) .&. 0xFFFF)

-- Result types

getLastError :: IO HRESULT
setLastError :: HRESULT -> IO ()


getLastError    = liftM hRESULT_FROM_WIN32 c_GetLastError
setLastError    = c_SetLastError . wIN32_FROM_HRESULT

-- Time

sleep        :: MILLISECONDS -> IO ()
getTickCount :: IO MILLISECONDS

queryPerformanceCounter   :: IO ULONGLONG
queryPerformanceFrequency :: IO ULONGLONG
performanceFrequency      :: ULONGLONG


sleep        = c_Sleep
getTickCount = c_GetTickCount

queryPerformanceCounter   = alloca $ \p -> c_QueryPerformanceCounter   p >> peek p
queryPerformanceFrequency = alloca $ \p -> c_QueryPerformanceFrequency p >> peek p
performanceFrequency      = unsafePerformIO queryPerformanceFrequency -- Safe because it's a constant throughout the execution of the program.

-- Object handles

-- Modules

getModuleHandle :: String -> IO HINSTANCE


getModuleHandle s = withMaybePSTR s $ \pstr -> c_GetModuleHandle pstr

-- Cursors

loadCursor    :: HINSTANCE -> String   -> IO HCURSOR
loadIdCursor  :: HINSTANCE -> IDCURSOR -> IO HCURSOR
loadSysCursor ::              IDCURSOR -> IO HCURSOR


loadCursor hInst = flip withPSTR $ c_LoadCursor hInst

loadIdCursor hInst = c_LoadCursor hInst . castToPtr
loadSysCursor      = loadIdCursor nullHINSTANCE

-- Message boxes

systemMessageBox :: String -> String -> MBTYPE -> IO INT


systemMessageBox text caption typ =
    withPSTR text    $ \c_Text  -> do
    withPSTR caption $ \c_Caption -> do
        c_SystemMessageBox nullPtr c_Text c_Caption typ

-- I/O handles

-- Strings

allocaPSTR :: Integral i => i -> (PSTR -> IO a) -> IO a

withPSTR         ::               String -> (PSTR -> IO a) -> IO a
withPSTRLen      :: Integral l => String -> ((PSTR, l) -> IO a) -> IO a
withMaybePSTR    ::               String -> (PSTR -> IO a) -> IO a
withMaybePSTRLen :: Integral l => String -> ((PSTR, l) -> IO a) -> IO a

peekPSTR         ::               PSTR -> IO String
peekPSTRLen      :: Integral l => (PSTR, l) -> IO String
peekMaybePSTR    ::               PSTR -> IO String
peekMaybePSTRLen :: Integral l => (PSTR, l) -> IO String


allocaPSTR len f = allocaArray (fromIntegral len + 1) f

withPSTR = withCWString

withPSTRLen s f = withCWStringLen s (\(p, l) -> f (p, fromIntegral l))

withMaybePSTR "" f = f nullPSTR
withMaybePSTR s  f = withCWString s f

withMaybePSTRLen "" f = f (nullPSTR, 0)
withMaybePSTRLen s  f = withCWStringLen s (\(p, l) -> f (p, fromIntegral l))

peekPSTR = peekCWString

peekPSTRLen (p, l) = peekCWStringLen (p, fromIntegral l)

peekMaybePSTR p =
    if p == nullPtr then
            return ""
        else
            peekCWString p

peekMaybePSTRLen (p, l) =
    if p == nullPtr then
            return ""
        else
            peekCWStringLen (p, fromIntegral l)
