module Win32.Window.API where

import Win32.Window.Types
import Win32.Window.Constants
import Win32.Window.Foreign

import Win32.GDI

import Win32.System

import Data.Bits
import Data.IORef
import Control.Monad
import Foreign.Marshal.Alloc
import Foreign.Marshal.Utils
import Foreign.Ptr
import Foreign.Storable
import System.IO.Unsafe



-- Window class

allocaClassName :: WndClass -> (PSTR -> HINSTANCE -> IO a) -> IO a

registerClass ::
    CSTYLE       ->  -- style
    WNDPROC      ->  -- lpfnWndProc
    INT          ->  -- cbClsExtra
    INT          ->  -- cbWndExtra
    HINSTANCE    ->  -- hInstance
    HICON        ->  -- hIcon
    HCURSOR      ->  -- hCursor
    HBRUSH       ->  -- hbrBackground
    String       ->  -- lpszMenuName
    String       ->  -- lpszClassName
    HICON        ->  -- hIconSm
    IO WndClass

unregisterClass :: WndClass -> IO BOOL

withRegisterClass ::
    CSTYLE       ->  -- style
    WNDPROC      ->  -- lpfnWndProc
    INT          ->  -- cbClsExtra
    INT          ->  -- cbWndExtra
    HINSTANCE    ->  -- hInstance
    HICON        ->  -- hIcon
    HCURSOR      ->  -- hCursor
    HBRUSH       ->  -- hbrBackground
    String       ->  -- lpszMenuName
    String       ->  -- lpszClassName
    HICON        ->  -- hIconSm
    (WndClass -> IO a) -> IO a

allocaClassName (WndClassName s i)         f = withPSTR s (\n -> f n i)
allocaClassName (WndClassRegistered a i _) f = f (nullPSTR `plusPtr` fromEnum a) i

registerClass style lpfnWndProc cbClsExtra cbWndExtra hInstance hIcon hCursor hbrBackground lpszMenuName lpszClassName hIconSm =
    allocaBytes   48            $ \wndClass ->
    withMaybePSTR lpszMenuName  $ \menuName ->
    withPSTR      lpszClassName $ \className -> do
        classProc <- mkWNDPROC lpfnWndProc
        pokeByteOff wndClass  0 (48 :: UINT)
        pokeByteOff wndClass  4 style
        pokeByteOff wndClass  8 classProc
        pokeByteOff wndClass 12 cbClsExtra
        pokeByteOff wndClass 16 cbWndExtra
        pokeByteOff wndClass 20 hInstance
        pokeByteOff wndClass 24 hIcon
        pokeByteOff wndClass 28 hCursor
        pokeByteOff wndClass 32 hbrBackground
        pokeByteOff wndClass 36 menuName
        pokeByteOff wndClass 40 className
        pokeByteOff wndClass 44 hIconSm
        atom <- c_RegisterClassEx wndClass
        return $ WndClassRegistered atom hInstance $ do
            freeHaskellFunPtr classProc

unregisterClass c@(WndClassRegistered a i f) =
    allocaClassName c $ \c_classname _ -> do
        result <- c_UnregisterClass c_classname i
        f
        return result

withRegisterClass style lpfnWndProc cbClsExtra cbWndExtra hInstance hIcon hCursor hbrBackground lpszMenuName lpszClassName hIconSm action = do
    wndClass <- registerClass style lpfnWndProc cbClsExtra cbWndExtra hInstance hIcon hCursor hbrBackground lpszMenuName lpszClassName hIconSm
    result <- action wndClass
    unregisterClass wndClass
    return result

-- Window

createWindow ::
    WndClass  ->  -- class
    String    ->  -- windowName
    WSTYLE    ->  -- dwStyle
    RECT      ->  -- rect
    HWND      ->  -- parentWindow
    HMENU     ->  -- menu
    IO HWND

createChildWindow ::
    WndClass  ->  -- class
    String    ->  -- windowName
    WSTYLE    ->  -- dwStyle
    RECT      ->  -- rect
    HWND      ->  -- parentWindow
    CHILDID   ->  -- child ID
    IO HWND

destroyWindow :: HWND -> IO BOOL
isWindow      :: HWND -> IO BOOL

withCreateUniqueWindow ::
    String         -> -- className
    CSTYLE         -> -- cStyle
    String         -> -- windowTitle
    WNDPROC        -> -- lpfnWndProc
    WSTYLE         -> -- wStyle
    RECT           -> -- rect
    HWND           -> -- parentWindow
    HMENU          -> -- menu
    HICON          -> -- hIcon
    HICON          -> -- hIconSm
    HCURSOR        -> -- hCursor
    HBRUSH         -> -- hbrBackground
    HINSTANCE      -> -- instance
    (HWND -> IO a) -> -- action
    IO a

createWindow class_ windowTitle dwStyle (RECT l t r b) parentWindow menu =
    allocaClassName class_ $ \c_class hInst ->
    withPSTR windowTitle   $ \c_windowTitle -> do

        hwnd <- c_CreateWindowEx
            (exstyleFromWSTYLE dwStyle)
            c_class
            c_windowTitle
            (styleFromWSTYLE dwStyle)
            l t (r-l) (b-t)
            parentWindow
            menu
            hInst
            nullPtr
        when (hwnd == nullHWND) $ do
            err <- getLastError
            putStrLn $ "Window creation failed (" ++ show err ++ ")"
        return hwnd

createChildWindow class_ windowTitle dwStyle rect parentWindow id =
    createWindow class_ windowTitle (dwStyle .|. wS_CHILD) rect parentWindow (HMENU $ nullPtr `plusPtr` fromIntegral id)

destroyWindow = c_DestroyWindow
isWindow      = c_IsWindow

withCreateUniqueWindow className cStyle windowTitle lpfnWndProc wStyle rect parentWindow menu hIcon hIconSm hCursor hbrBackground hInst action =
    withRegisterClass cStyle lpfnWndProc 0 0 hInst hIcon hCursor hbrBackground "" className hIconSm $ \wndClass -> do
        hwnd <- createWindow wndClass windowTitle wStyle rect parentWindow menu
        result <- action hwnd
        validWnd <- isWindow hwnd
        when validWnd $ destroyWindow hwnd >> return ()
        return result

-- Utils

adjustWindowRectEx :: RECT -> WSTYLE -> BOOL -> Maybe RECT

adjustWindowRectEx rect wstyle menu = unsafePerformIO $
    alloca $ \prect -> do
        poke prect rect
        result <- c_AdjustWindowRectEx prect (styleFromWSTYLE wstyle) menu (exstyleFromWSTYLE wstyle)
        if result then (liftM Just $ peek prect) else return Nothing

-- Properties

getWindowProc  :: HWND -> IO WNDPROC
setWindowProc  :: HWND -> WNDPROC -> IO WNDPROC
subclassWindow :: HWND -> (WNDPROC -> WNDPROC) -> IO WNDPROC

getWindowStyle :: HWND -> IO WSTYLE
setWindowStyle :: HWND -> WSTYLE -> IO WSTYLE

getWindowText       :: HWND -> IO String
getWindowTextLength :: HWND -> IO INT
setWindowText       :: HWND -> String -> IO BOOL

getClientRect :: HWND -> IO RECT

getWindowFont :: HWND -> IO HFONT
setWindowFont :: HWND -> HFONT -> BOOL -> IO ()

getDC     :: HWND -> IO HDC
releaseDC :: HWND -> HDC -> IO BOOL
withGetDC :: HWND -> (HDC -> IO a) -> IO (Maybe a)

getWindowProc hwnd      = liftM (getWNDPROC . castToFuncPtr) $ c_GetWindowDWORD hwnd gWL_WNDPROC
setWindowProc hwnd proc = do
    ioRef <- newIORef $ return ()
    newProc <- mkWNDPROC $ finalizer ioRef
    writeIORef ioRef $ freeHaskellFunPtr newProc
    liftM (getWNDPROC . castToFuncPtr) $ c_SetWindowDWORD hwnd gWL_WNDPROC $ castFromFuncPtr newProc
    where
    finalizer ioRef hwnd msg wParam lParam
        | msg == wM_NCDESTROY = do
            f <- readIORef ioRef
            f
            proc hwnd msg wParam lParam
        | otherwise =
            proc hwnd msg wParam lParam

subclassWindow hwnd proc = do
    oldProc <- getWindowProc hwnd
    setWindowProc hwnd (proc oldProc)

getWindowStyle hwnd = do
    style <- c_GetWindowDWORD hwnd gWL_STYLE
    exstyle <- c_GetWindowDWORD hwnd gWL_EXSTYLE
    return $ (fromIntegral exstyle `shiftL` 32) + fromIntegral style

setWindowStyle hwnd wstyle = do
    style <- c_SetWindowDWORD hwnd gWL_STYLE (styleFromWSTYLE wstyle)
    exstyle <- c_SetWindowDWORD hwnd gWL_EXSTYLE (exstyleFromWSTYLE wstyle)
    return $ (fromIntegral exstyle `shiftL` 32) + fromIntegral style

getWindowText hwnd = do
    length <- getWindowTextLength hwnd
    allocaPSTR length $ \ptr -> do
        b <- c_GetWindowText hwnd ptr (length + 1)
        if b then peekPSTR ptr
             else return ""

getWindowTextLength = c_GetWindowTextLength

setWindowText hwnd text = withPSTR text $ \c_text -> c_SetWindowText hwnd c_text

getClientRect hwnd = alloca $ \prect -> do
    result <- c_GetClientRect hwnd prect
    if result then peek prect
              else return $ RECT 0 0 0 0

getWindowFont hwnd              = liftM castToHFONT $ sendMessage hwnd wM_GETFONT 0 0
setWindowFont hwnd font redraw  = sendMessage hwnd wM_SETFONT (castFromHFONT font) (fromBool redraw) >> return ()

getDC     = c_GetDC
releaseDC = c_ReleaseDC
withGetDC hwnd a = do
    dc <- getDC hwnd
    if dc /= nullHDC then do
            result <- a dc
            releaseDC hwnd dc
            return $ Just result
        else
            return Nothing

-- Actions

sendMessage :: HWND -> WMSG -> WPARAM -> LPARAM -> IO LRESULT
postMessage :: HWND -> WMSG -> WPARAM -> LPARAM -> IO BOOL

invalidateRect :: HWND -> Maybe RECT -> BOOL -> IO BOOL

sendMessage = c_SendMessage
postMessage = c_PostMessage

invalidateRect hwnd Nothing erase = c_InvalidateRect hwnd nullPtr erase
invalidateRect hwnd (Just r) erase = alloca $ \rect -> do
    poke rect r
    c_InvalidateRect hwnd rect erase

-- Window manager

showWindow   :: HWND -> SW -> IO BOOL
setWindowPos :: HWND -> HWND -> RECT -> SWP -> IO BOOL

getFocus     :: IO HWND
setFocus     :: HWND -> IO HWND

showWindow = c_ShowWindow

setWindowPos hwnd hwndAfter rect flags = c_SetWindowPos
    hwnd hwndAfter
    (getRECTLeft   rect)
    (getRECTTop    rect)
    (getRECTWidth  rect)
    (getRECTHeight rect)
    flags

getFocus = c_GetFocus
setFocus = c_SetFocus

-- Message boxes

messageBox hwnd text caption typ =
    withPSTR text    $ \c_Text  -> do
    withPSTR caption $ \c_Caption -> do
        c_MessageBox hwnd c_Text c_Caption typ
